/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.ast;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;

import javax.xml.bind.annotation.XmlAttribute;
import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashSet;

/**
 * Pair of values.
 *
 * <p>Object can be treated as a sub-list of indices.
 * One (low) value is less or equal than another (high) value.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 */
public class Range {

  @XmlAttribute private final Node low;

  @XmlAttribute private final Node high;

  /**
   * Default constructor.
   *
   * @param low low index
   * @param high high index
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when any of the arguments is {@code null};
   *         (2) when the second argument is always less than first argument.
   */
  public Range(final Node low, final Node high) {
    InvariantChecks.checkNotNull(low);
    InvariantChecks.checkNotNull(high);

    this.low = low;
    this.high = high;
  }

  /**
   * Creates range with the specified low and high indices.
   * @param low The smaller (low) index value.
   * @param high The bigger (high) index value.
   */
  public Range(final int low, final int high) {
    this(NodeValue.newInteger(low), NodeValue.newInteger(high));
  }

  /**
   * Constructs a copy of the specified range.
   * @param integerRange The integer range.
   */
  public Range(final IntegerRange integerRange) {
    this(integerRange.getMin().intValue(), integerRange.getMax().intValue());
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return a full copy of the current object
   */
  public Range deepCopy() {
    return new Range(this.low.deepCopy(), this.high.deepCopy());
  }

  /**
   * Returns low index.
   *
   * @return low index
   */
  public Node getLow() {
    return this.low;
  }

  /**
   * Returns high index.
   *
   * @return high index
   */
  public Node getHigh() {
    return this.high;
  }

  /**
   * Returns variables that are used in this object.
   * @return The collection of variables that are used in this object.
   */
  public Collection<NodeVariable> getVariables() {

    final Collection<NodeVariable> variables = new LinkedHashSet<>();

    variables.addAll(ExprUtils.getVariables(getHigh()));
    variables.addAll(ExprUtils.getVariables(getLow()));

    return variables;
  }

  /**
   * Checks whether this range has constant borders.
   * @return {@code true} when this range has constant borders, {@code false} otherwise.
   */
  public boolean isConstantRange() {

    return ExprUtils.isKind(Node.Kind.VALUE, getHigh(), getLow());
  }

  /**
   * Checks whether this object is a range of integer values.
   * @return {@code true} if this object is integer value range, {@code false} otherwise.
   */
  public boolean isIntegerRange() {
    return getLow().isType(DataType.INTEGER) && getLow().getKind() == Node.Kind.VALUE
        && getHigh().isType(DataType.INTEGER) && getHigh().getKind() == Node.Kind.VALUE;
  }

  /**
   * Converts this object to integer range, if fails, returns {@code null}.
   * @return the identical integer range of {@code null}, if fails.
   */
  public IntegerRange toIntegerRange() {

    return new IntegerRange(toBigInt(getLow()), toBigInt(getHigh()));
  }

  private BigInteger toBigInt(final Node node) {

    return isIntegerRange() ? toInt(node) : toBigInteger(node);
  }

  private static BigInteger toBigInteger(final Node node) {

    final NodeValue toInt =
        (NodeValue) TypeConversion.coerce(Reducer.reduce(node), DataType.INTEGER);

    InvariantChecks.checkNotNull(toInt, "Can't cast to int: " + node);

    return toInt.getInteger();
  }

  private static BigInteger toInt(final Node node) {

    return ((NodeValue)node).getInteger();
  }

  @Override
  public String toString() {

    return String.format("%s:%s", getLow(), getHigh());
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }

    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final Range range = (Range) obj;

    return this.low.equals(range.low) && this.high.equals(range.high);
  }

  @Override
  public int hashCode() {
    int result = this.low.hashCode();
    result = 31 * result + this.high.hashCode();
    return result;
  }
}
