/*
 * Copyright 2012-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.ast.basis;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.LinkedHashMap;
import java.util.Map;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link SymbolTable} implements a simple symbol table based on Java hash tables.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class SymbolTable<Tag extends Enum<Tag>> {
  /** Maps a name to a node (an entity descriptor). */
  private final Map<String, SyntaxNode<Tag>> table = new LinkedHashMap<>();
  /** Maps a tag to a linked list of all nodes with that tag. */
  private final Map<Tag, Collection<SyntaxNode<Tag>>> slices = new LinkedHashMap<>();
  /** Contains all children of the node. */
  private final Collection<SyntaxNode<Tag>> children = new ArrayList<>();

  /** Symbol table of the parent scope. */
  private SymbolTable<Tag> parent;

  /** Default redefinition handler. */
  private ErrorHandler<Tag> redefinition = new ErrorHandler<Tag>() {
    @Override
    public void handle(final SyntaxNode<Tag> newNode, final SyntaxNode<Tag> oldNode) {
      throw new IllegalStateException(
          String.format("The symbol table already contains the name '%s'", newNode.getName()));
    }
  };

  /**
   * Returns the parent symbol table.
   *
   * @return the parent symbol table or {@code null} if the table is a top-level one.
   */
  protected SymbolTable<Tag> getParent() {
    return parent;
  }

  /**
   * Sets the parent symbol table.
   *
   * @param parent the symbol table to be set.
   */
  protected void setParent(final SymbolTable<Tag> parent) {
    InvariantChecks.checkNotNull(parent);
    this.parent = parent;
  }

  /**
   * Returns the collection of items.
   *
   * @return the items.
   */
  public Collection<SyntaxNode<Tag>> items() {
    return children;
  }

  /**
   * Returns the collection of items with the given tag.
   *
   * @param tag the tag.
   * @return the items.
   */
  public Collection<SyntaxNode<Tag>> items(final Tag tag) {
    InvariantChecks.checkNotNull(tag);

    Collection<SyntaxNode<Tag>> slice = slices.get(tag);
    if (slice == null) {
      slice = new ArrayList<>();
    }

    return slice;
  }

  /**
   * Returns the collection of items whose tags belong to the given set.
   *
   * @param tags the set of tags.
   * @return the items.
   */
  public Collection<SyntaxNode<Tag>> items(final EnumSet<Tag> tags) {
    InvariantChecks.checkNotNull(tags);

    final Collection<SyntaxNode<Tag>> selection = new ArrayList<>();

    for (final Tag tag : tags) {
      selection.addAll(items(tag));
    }

    return selection;
  }

  /**
   * Returns an item of the symbol table with the given name.
   *
   * @param name the item name.
   * @return an item with the given name or {@code null} if the symbol table does not contain such
   *         an item.
   */
  public SyntaxNode<Tag> find(final String name) {
    InvariantChecks.checkNotNull(name);
    return table.get(name);
  }

  /**
   * Returns an item of the symbol table (or the parent tables) with the given name
   *
   * @param name the item name.
   * @return an item with the given name or {@code null} if the symbol table and the parent tables
   *         do not contain such an item.
   */
  public final SyntaxNode<Tag> findRecursively(final String name) {
    InvariantChecks.checkNotNull(name);

    for (SymbolTable<Tag> table = this; table != null; table = table.getParent()) {
      final SyntaxNode<Tag> child = table.find(name);

      if (child != null) {
        return child;
      }
    }

    return null;
  }

  /**
   * Adds an item into the symbol table.
   *
   * @param item the item to be added.
   */
  public void add(final SyntaxNode<Tag> item) {
    InvariantChecks.checkNotNull(item);

    final Tag tag = item.getTag();
    final String name = item.getName();

    // If the item has a name, add the item into the symbol table.
    if (item.hasName() && name != null) {
      if (table.containsKey(name)) {
        // Invoke the redefinition callback.
        redefinition.handle(item, table.get(name));
      }

      table.put(name, item);
    }

    // Add the item into the slice.
    Collection<SyntaxNode<Tag>> slice = slices.get(tag);
    if (slice == null) {
      slice = new ArrayList<>();
    }

    slice.add(item);
    slices.put(tag, slice);

    // Add the item into the list of children.
    children.add(item);
  }

  /**
   * Adds items into the symbol table.
   *
   * @param items the items to be added.
   */
  public final void addAll(final Collection<SyntaxNode<Tag>> items) {
    InvariantChecks.checkNotNull(items);

    for (final SyntaxNode<Tag> item : items) {
      add(item);
    }
  }

  /**
   * Sets the redefinition handler.
   * 
   * @param handler the handler to be set.
   */
  public void setRedefinitionHandler(final ErrorHandler<Tag> handler) {
    InvariantChecks.checkNotNull(handler);
    this.redefinition = handler;
  }
}
