/*
 * Copyright 2012-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.ast.basis;

import java.util.Collection;
import java.util.EnumSet;
import java.util.LinkedHashMap;
import java.util.Map;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link SyntaxNode} represents a basic node of an abstract syntax tree (AST).
 *
 * @see SymbolTable
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class SyntaxNode<Tag extends Enum<Tag>> implements Cloneable {
  /**
   * {@link Kind} classifies nodes according to the has-name and has-scope flags.
   */
  public enum Kind {
    /** Node with name and scope (module, function, etc.). */
    CLASS(true, true),
    /** Node with name but without scope (instance, variable, etc.). */
    OBJECT(true, false),
    /** Node with scope but without name (block, loop body, etc.). */
    BLOCK(false, true),
    /** Node without name and scope (statement, expression, etc.). */
    BASIC(false, false);

    private final boolean hasName;
    private final boolean hasScope;

    Kind(final boolean hasName, final boolean hasScope) {
      this.hasName = hasName;
      this.hasScope = hasScope;
    }

    public boolean hasName() {
      return hasName;
    }

    public boolean hasScope() {
      return hasScope;
    }
  }

  /** Tag class. */
  private final Class<Tag> tagType;
  /** Tag of the node (not {@code null}). */
  private final Tag tag;
  /** Admissible tags of the node's children (not {@code null}). */
  private final EnumSet<Tag> childrenTags;
  /** Kind of the node (not {@code null}). */
  private final Kind kind;

  /** Meta Information. */
  private final Map<String, Object> attributes;

  /** Name of the node ({@code null} if the node has no name). */
  private String name;
  /** Parent node ({@code null} for the root node). */
  private SyntaxNode<Tag> parent;
  /** Symbol table of the node ({@code null} if the node has no scope). */
  private SymbolTable<Tag> symbolTable;

  /**
   * Creates an AST node.
   *
   * @param tagType the node tag type.
   * @param tag the node tag.
   * @param childrenTags the admissible tags of the node's children.
   * @param kind the node kind.
   * @param name the node name.
   * @param parent the node parent.
   */
  protected SyntaxNode(
      final Class<Tag> tagType,
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final Kind kind,
      final String name,
      final SyntaxNode<Tag> parent) {
    InvariantChecks.checkNotNull(tagType);
    InvariantChecks.checkNotNull(tag);
    InvariantChecks.checkNotNull(childrenTags);
    InvariantChecks.checkNotNull(kind);

    this.tagType = tagType;
    this.tag = tag;
    this.childrenTags = childrenTags;
    this.kind = kind;

    this.attributes = new LinkedHashMap<>();

    this.name = name;

    // Even a node with no scope has the symbol table (as soon as the node's parent becomes known,
    // all of the symbols are moved to the parent's symbol table).
    this.symbolTable = new SymbolTable<>();

    // Parent should be set after the symbol table is constructed.
    if (parent != null) {
      setParent(parent);
    }
  }

  protected SyntaxNode(
      final Class<Tag> tagType,
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final Kind kind,
      final SyntaxNode<Tag> parent) {
    this(tagType, tag, childrenTags, kind, null, parent);
  }

  protected SyntaxNode(
      final Class<Tag> tagType,
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final Kind kind) {
    this(tagType, tag, childrenTags, kind, null);
  }

  protected SyntaxNode(
      final Class<Tag> tagType,
      final Tag tag,
      final Kind kind) {
    this(tagType, tag, EnumSet.noneOf(tagType), kind, null);
  }

  protected SyntaxNode(final SyntaxNode<Tag> r) {
    InvariantChecks.checkNotNull(r);

    this.tagType = r.tagType;
    this.tag = r.tag;
    this.childrenTags = r.childrenTags;
    this.kind = r.kind;
    this.attributes = new LinkedHashMap<>(r.attributes);
    this.name = r.name;
    this.parent = r.parent;
    this.symbolTable = r.symbolTable;
  }

  /**
   * Returns the type of the node tag.
   *
   * @return the node tag type.
   */
  public final Class<Tag> getTagType() {
    return tagType;
  }

  /**
   * Returns the tag of the node.
   *
   * @return the node tag.
   */
  public final Tag getTag() {
    return tag;
  }

  /**
   * Returns the kind of the node.
   *
   * @return the node kind.
   */
  public final Kind getKind() {
    return kind;
  }

  /**
   * Checks whether the node contains the given attribute.
   *
   * @param key the attribute key.
   * @return {@code true} if the node contains the attribute; {@code false} otherwise.
   */
  public final boolean containsAttribute(final String key) {
    InvariantChecks.checkNotNull(key);
    return attributes.containsKey(key);
  }

  /**
   * Returns the attribute value.
   *
   * @param key the attribute key.
   * @return the attribute value or {@code null} if the attribute has not been set.
   */
  public final Object getAttribute(final String key) {
    InvariantChecks.checkNotNull(key);
    return attributes.get(key);
  }

  /**
   * Sets the attribute value.
   *
   * @param key the attribute key.
   * @param value the attribute value
   */
  public final void setAttribute(final String key, final Object value) {
    InvariantChecks.checkNotNull(key);
    InvariantChecks.checkNotNull(value);

    attributes.put(key, value);
  }

  /**
   * Sets the attribute values.
   *
   * @param attributes the attributes to be set.
   */
  public final void setAttributes(final Map<String, Object> attributes) {
    InvariantChecks.checkNotNull(attributes);

    this.attributes.putAll(attributes);
  }

  /**
   * Checks whether the node has name.
   *
   * @return {@code true} if the node has name; {@code false} otherwise.
   */
  public final boolean hasName() {
    return kind.hasName();
  }

  /**
   * Checks whether the node has scope.
   *
   * @return {@code true} if the node has scope; {@code false} otherwise.
   */
  public final boolean hasScope() {
    return kind.hasScope();
  }

  /**
   * Returns the name of the node.
   * 
   * @return the node name.
   */
  public final String getName() {
    return name;
  }

  /**
   * Sets the name of the node.
   *
   * @param name the node name.
   */
  public final void setName(final String name) {
    InvariantChecks.checkNotNull(name);
    this.name = name;
  }

  /**
   * Returns the parent of the node or {@code null} if the node is a root.
   * 
   * @return the node parent.
   */
  public final SyntaxNode<Tag> getParent() {
    return parent;
  }

  /**
   * Sets the parent of the node.
   * 
   * @param parent the node parent.
   */
  public final void setParent(final SyntaxNode<Tag> parent) {
    InvariantChecks.checkNotNull(parent);

    this.parent = parent;
    this.symbolTable.setParent(parent.symbolTable);
  }

  /**
   * Returns the collection of the children.
   *
   * @return the children.
   */
  public final Collection<SyntaxNode<Tag>> getChildren() {
    return symbolTable.items();
  }

  /**
   * Returns the collection of the children of the given tag.
   *
   * @param tag the child tag.
   * @return the children.
   */
  public final Collection<SyntaxNode<Tag>> getChildren(final Tag tag) {
    InvariantChecks.checkNotNull(tag);
    return symbolTable.items(tag);
  }

  /**
   * Searches for a child with the given name in the current scope.
   *
   * @param name the child name.
   * @return the child node or {@code null}.
   */
  public final SyntaxNode<Tag> find(final String name) {
    InvariantChecks.checkNotNull(name);
    return symbolTable.find(name);
  }

  /**
   * Searches for a child with the given name in the current and upper scopes.
   *
   * @param name the child name.
   * @return the child node or {@code null}.
   */
  public final SyntaxNode<Tag> findRecursively(final String name) {
    InvariantChecks.checkNotNull(name);
    return symbolTable.findRecursively(name);
  }

  /**
   * Adds a child to the node.
   *
   * @param child the child to be added.
   */
  public final void add(final SyntaxNode<Tag> child) {
    InvariantChecks.checkNotNull(child);

    // Check admissibility of the child's tag.
    InvariantChecks.checkTrue(childrenTags.contains(child.getTag()),
        String.format("The child of '%s' has the illegal tag '%s'", getTag(), child.getTag()));

    // Set the child's parent.
    final SyntaxNode<Tag> parent = child.getParent();
    InvariantChecks.checkTrue(parent == null || parent == this, "Cannot change the parent node");

    child.setParent(this);

    // Even unnamed children are put into the symbol table.
    symbolTable.add(child);
  }

  /**
   * Sets the redefinition handler.
   * 
   * @param handler the handler to be set.
   */
  protected final void setRedefinitionHandler(final ErrorHandler<Tag> handler) {
    InvariantChecks.checkNotNull(handler);
    symbolTable.setRedefinitionHandler(handler);
  }
}
